/*
 * fserver.c - Simple File Server
 *
 * Copyright (c) 2000 Cogito, LLC 
 *
 * Cogito grants you ("Licensee") a non-exclusive, royalty free, license to
 * use, modify and redistribute this software in source and binary form,
 * provided that i) this copyright notice and license appear on all copies
 * of the software; and ii) Licensee does not utilize the software in a
 * manner which is disparaging to Cogito.
 *
 * This software is provided "AS IS," without warranty of any kind.
 *
 * 12/2000: Initial version.
 *          M. Tim Jones <mtj@cogitollc.com>
 * 12/28/2000: Added a SIGCHLD handler to avoid zombies...
 *          M. Tim Jones <mtj@cogitollc.com>
 *
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <errno.h>
#include <sys/un.h>
#include <unistd.h>

#define SERV_PORT	15000
#define BUF_SIZE        256
#define VERSION         "0.02"
#define BUILD_DATE      __DATE__
#define BUILD_TIME      __TIME__

/* 
 * Signal handler for SIGPIPE... 
 */
void ouch ( int sigNum )
{
  // Just avoid crashing the server with EPIPE...
}


/* 
 * Signal handler for SIGHUP... 
 */
void sighup ( int sigNum )
{
  if (readConfigFile() == -1) {
    printf("ERROR in configuration file...\n");
  }
}


/* 
 * Signal handler for SIGCHLD... 
 */
void sigchld ( int sigNum )
{
  pid_t pid;

  while ((pid = waitpid((pid_t)-1, (int *)0, WNOHANG)) > 0);

}


/* Once a socket connection is established, converse with the client to
 * identify the request.  Once the request is identified, call the particular
 * handler function to take it from there.
 */

int handleConnection(int sockfd, char *subdir)
{
  char line[BUF_SIZE+1];
  int  ret;
  int  retStatus = 0;

  /* The socket is open from the client, the first newline terminated line
   * contains the command to be performed and drives the rest of the
   * interaction with the client.
   */

  memset(line, 0, BUF_SIZE+1);
  ret = read(sockfd, line, BUF_SIZE);

  if (ret <= 0) {
    printf("Bad socket connection (%d)\n", errno);
    return;
  }

  if (!strncmp(line, "$ls", 3)) {
    handleRemoteList(sockfd, &line[4], subdir);
  } else if (!strncmp(line, "$put", 4)) {
    handleRemotePut(sockfd, &line[5]);
  } else if (!strncmp(line, "$get", 4)) {
    handleRemoteGet(sockfd, &line[5], subdir);
  } else if (!strncmp(line, "$shutdown", 9)) {
    /* TODO: Not yet implemented... */
    retStatus = -1;
  } else {
    printf("Bad command from client - %s\n", line);
  }

  return(retStatus);
}


/* Main entry for the simple file server.  Create a simple server socket
 * and wait for requests.  Each request is forked off to a child for
 * further handling.
 */

int main(int argc, char *argv[])
{
  int listenfd, connfd;
  pid_t childpid;
  socklen_t clilen;
  struct sockaddr_in cliaddr, servaddr;
  const int on = 1;

  extern int authenticateClient(char *);
  extern char *getSubdir(char *);

  printf("\nfserver starting - V%s %s/%s\n\n", VERSION, BUILD_DATE, BUILD_TIME);

  (void)signal(SIGPIPE, ouch);
  (void)signal(SIGHUP, sighup);
  (void)signal(SIGCHLD, sigchld);

  /* Read the config file
   */
  if (readConfigFile() == -1) {
    exit(0);
  }

  listenfd = socket(AF_INET, SOCK_STREAM, 0);

  /* Allow us to bind to the port even if it's specified as in use. */
  setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));

  bzero(&servaddr, sizeof(servaddr));
  servaddr.sin_family = AF_INET;
  servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
  servaddr.sin_port = htons(SERV_PORT);

  bind(listenfd, (struct sockaddr *)&servaddr, sizeof(servaddr));

  listen(listenfd, 5);

  while (1) {

    clilen = sizeof(cliaddr);
    connfd = accept(listenfd, (struct sockaddr *)&cliaddr, &clilen);

    printf("Connection from %s (base path %s)\n", 
            inet_ntoa(cliaddr.sin_addr),
            getSubdir(inet_ntoa(cliaddr.sin_addr)));

    /* Fork off a child to handle each individual connection.
     */
    if ( (childpid = fork()) == 0) {
      close(listenfd);
      if (authenticateClient(inet_ntoa(cliaddr.sin_addr)) > 0) {
        handleConnection(connfd, getSubdir(inet_ntoa(cliaddr.sin_addr)));
      } else {
        printf("------DENIED------\n");
      }
      waitpid(childpid, (int *)0, WNOHANG);
      exit(0);
    }

    close(connfd);

  }

  close(listenfd);

}
